"""
Events & Subscriptions
======================

The workflow for using events is this:

1. Subscribe to an event publisher, using :func:`subscribe`
    You can choose publishers (and exclusions) from the information that
    :func:`list_publisher` returns.

    Note the subscription_id, that :func:`subscribe` returns to you
2. Now call :func:`listen` in a while loop.
    The request will be open for as long as keep_alive timeout is set or an
    event happens.
    If the timeout is reached, re-send the request.
    When you receive an event, you can act upon it and re-send the request.
3. You can always add or remove subscriptions/exclusions with your
    subscription_id.
    You can also change the poll and keep-alive timeouts.
4. When your done, call :func:`unsubscribe`.
    If you do not do that, JDownloader will delete the subscription after some
    time has passed.

.. note:: If you do not want to have a loss of actuality in your application,
    subscribe _before_ you load your data. Otherwise it can happen, that there
    has been a change of content between you loading the data and then
    subscribing (especially if done over network).

"""

from .jd_types import SubscriptionResponse, PublisherResponse
from typing import Any, List


class Events:
    def __init__(self, device):
        self.device = device
        self.endpoint = "events"

    def action(self, route: str, params: Any = None) -> Any:
        route = f"/{self.endpoint}{route}"
        return self.device.connection_helper.action(route, params)

    def add_subscription(
        self,
        subscription_id: int,
        subscriptions: List[str] = [],
        exclusions: List[str] = [],
    ) -> SubscriptionResponse:
        """Add subscriptions/exclusions to an existing Subscription.

        :param subscription_id: Subscription ID, handed out by
            :func:`subscribe`
        :type subscription_id: int
        :param subscriptions: A list of event publishers
        :type subscriptions: List[str]
        :param exclusions: A list of excluded events
        :type exclusions: List[str]
        :returns: A subscription object
        :rtype: Subscription
        """

        params = [subscription_id, subscriptions, exclusions]
        resp = self.action("/addsubscription", params)
        subscription_response = SubscriptionResponse(**resp)
        return subscription_response

    def change_subscription_timeouts(
        self, subscription_id: int, poll_timeout: int, keep_alive: int
    ) -> SubscriptionResponse:
        """Change subscription timeouts for a subcription.

        :param subscription_id: Subscription ID, handed out by
            :func:`subscribe`
        :type subscription_id: int
        :param poll_timeout: Timeout after which a request closes, if there has
            been no event
        :type poll_timeout: int
        :param keep_alive: Timeout for keeping a subscription alive
        :type keep_alive: int
        :returns: A subscription object
        :rtype: Subscription
        """

        params = [subscription_id, poll_timeout, keep_alive]
        resp = self.action("/changesubscriptiontimeouts", params)
        subscription_response = SubscriptionResponse(**resp)
        return subscription_response

    def get_subscription(self, subscription_id: int) -> SubscriptionResponse:
        """Get a Subscription object by id.

        :returns: A subscription object
        :rtype: Subscription
        """

        params = [subscription_id]
        resp = self.action("/getsubscription", params)
        subscription_response = SubscriptionResponse(**resp)
        return subscription_response

    def listen(self, subscription_id: int) -> List[dict]:
        """Listen for events for a subscription.

        :param subscription_id: The id for a subscription
        :type subscription_id: int
        :returns: A list of events, or if the poll_timeout is reached, nothing.
        :rtype: List[dict]
        """

        params = [subscription_id]
        resp = self.action("/listen", params)
        return resp

    def list_publisher(self) -> List[PublisherResponse]:
        """List all event publishers and their events."""

        resp = self.action("/listpublisher")

        publisher_responses = []
        for response in resp:
            publisher_response = PublisherResponse(**response)
            publisher_responses.append(publisher_response)

        return publisher_responses

    def remove_subscription(
        self,
        subscription_id: int,
        subscriptions: List[str] = [],
        exclusions: List[str] = [],
    ) -> SubscriptionResponse:
        """Remove subscriptions/exclusions to an existing Subscription.

        :param subscription_id: Subscription ID, handed out by
            :func:`subscribe`
        :type subscription_id: int
        :param subscriptions: A list of event publishers
        :type subscriptions: List[str]
        :param exclusions: A list of excluded events
        :type exclusions: List[str]
        :returns: A subscription object
        :rtype: Subscription
        """

        params = [subscription_id, subscriptions, exclusions]
        resp = self.action("/removesubscription", params)
        subscription_response = SubscriptionResponse(**resp)
        return subscription_response

    def set_subscription(
        self,
        subscription_id: int,
        subscriptions: List[str] = [],
        exclusions: List[str] = [],
    ) -> SubscriptionResponse:
        """Set subscriptions/exclusions for an existing Subscription.

        :param subscription_id: Subscription ID, handed out by
            :func:`subscribe`
        :type subscription_id: int
        :param subscriptions: A list of event publishers
        :type subscriptions: List[str]
        :param exclusions: A list of excluded events
        :type exclusions: List[str]
        :returns: A subscription object
        :rtype: Subscription
        """

        params = [subscription_id, subscriptions, exclusions]
        resp = self.action("/setsubscription", params)
        subscription_response = SubscriptionResponse(**resp)
        return subscription_response

    def subscribe(
        self, subscriptions: List[str] = [], exclusions: List[str] = []
    ) -> SubscriptionResponse:
        """Create a new subscription.

        :param subscriptions: A list of event publishers
        :type subscriptions: List[str]
        :param exclusions: A list of excluded events
        :type exclusions: List[str]
        :returns: A subscription object
        :rtype: Subscription
        """

        params = [subscriptions, exclusions]
        resp = self.action("/subscribe", params)
        subscription_response = SubscriptionResponse(**resp)
        return subscription_response

    def unsubscribe(self, subscription_id: int) -> SubscriptionResponse:
        """Create a new subscription.

        :param subscription_id: ID of the subscription that is terminated
        :type subscription_id: int
        :returns: A empty subscription object
        :rtype: Subscription
        """

        params = [subscription_id]
        resp = self.action("/unsubscribe", params)
        subscription_response = SubscriptionResponse(**resp)
        return subscription_response



from .jd_types import DirectConnectionInfos
from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    from .jd_device import JDDevice


class Device:
    def __init__(self, device: "JDDevice"):

        self.device = device
        self.endpoint = "device"

    def action(self, route: str, params: Any = None) -> Any:
        route = f"/{self.endpoint}{route}"
        return self.device.connection_helper.action(route, params)

    def get_direct_connection_infos(self) -> DirectConnectionInfos:

        resp = self.action('/getDirectConnectionInfos')
        if resp:
            return DirectConnectionInfos(**resp)

        return None

    def get_session_public_key(self) -> str:

        resp = self.action('/getSessionPublicKey')
        return resp

    def ping(self) -> bool:

        resp = self.action('/ping')
        return resp



from .jd_types import LogFolder
from typing import Any, List


class Log:
    def __init__(self, device):
        self.device = device
        self.endpoint = "log"

    def action(self, route: str, params: Any = None) -> Any:
        route = f"/{self.endpoint}{route}"
        return self.device.connection_helper.action(route, params)

    def get_available_logs(self) -> List[LogFolder]:
        """Returns a list of available logs.

        :return: List of log folders
        :rtype: jd_types.LogFolder
        """

        resp = self.action("/getAvailableLogs")

        log_folders = []
        for folder in resp:
            log_folder = LogFolder(**folder)
            log_folders.append(log_folder)

        return log_folders

    def send_log_file(self, log_folders) -> str:
        """Returns a log file.

        :return: The log file
        :rtype: str
        """

        params = [log_folders]
        resp = self.action("/sendLogFile", params)
        return resp



from .jd_device import JDDevice
from .myjd_connection_helper import MyJDConnectionHelper
from Crypto.Cipher import AES
from typing import Any, Dict, List
from urllib.parse import quote
import base64
import hashlib
import hmac
import json
import requests
import time

BS = 16


def PAD(s: bytes) -> bytes:
    """Pad a string

    :param s: String to pad
    :type s: str
    :return: Padded ``s``.
    :rtype: str
    """

    return s + ((BS - len(s) % BS) * chr(BS - len(s) % BS)).encode()


def UNPAD(s: bytes) -> bytes:
    """Unpad a string.

    :param s: String to unpad
    :type s: str
    :return: Unpadded ``s``.
    :rtype: str
    """

    return s[0 : -s[-1]]


class MyJDConnector:
    """Main class for connecting to the MyJD API."""

    def __init__(self) -> None:
        """Initialize MyJD connector."""

        self.__request_id = int(time.time() * 1000)
        self.__api_url = "https://api.jdownloader.org"
        self.__app_key = "https://pglaum.srht.site/pyjd-api"
        self.__api_version = 1
        self.__devices = None

        self.__login_secret = None
        self.__device_secret = None
        self.__session_token = None
        self.__regain_token = None
        self.__server_encryption_token = None
        self.__device_encryption_token = None

        self.__connected = False

    def get_session_token(self) -> str:
        """Get the session token

        :return: Returns ``self.__session_token``.
        :rtype: str
        """

        return self.__session_token

    def is_connected(self) -> bool:
        """Indicate if a connection is established.

        :return: Returns ``self.__connected``.
        :rtype: bool
        """

        return self.__connected

    def set_app_key(self, app_key: str) -> None:
        """Set the app key.

        Use this function to set a custom app key for your own projects.

        :param app_key: The key to set.
        :type app_key: str
        """

        self.__app_key = app_key

    def __create_secret(self, email: str, password: str, domain: str) -> bytes:
        """Create the login_secret and device_secret.

        :param email: MyJDownloader user email
        :type email: str
        :param password: MyJDownloader user email
        :type password: str
        :param domain: The domain of the secret ("server" for login_secret and
            "device" for device_secret)
        :type domain: str
        :return: secret hash
        :rtype: str
        """

        secret_hash = hashlib.sha256()
        secret_hash.update(
            email.lower().encode("utf-8")
            + password.encode("utf-8")
            + domain.lower().encode("utf-8")
        )

        return secret_hash.digest()

    def __update_encryption_tokens(self) -> None:
        """Update the ``server_encryption_token`` and
        ``device_encryption_token``.
        """

        if self.__server_encryption_token is None:
            old_token = self.__login_secret
        else:
            old_token = self.__server_encryption_token
        new_token = hashlib.sha256()
        new_token.update(old_token + bytearray.fromhex(self.__session_token))
        self.__server_encryption_token = new_token.digest()
        new_token = hashlib.sha256()
        new_token.update(self.__device_secret + bytearray.fromhex(self.__session_token))
        self.__device_encryption_token = new_token.digest()

    def __create_signature(self, key: str, data: str) -> bytes:
        """Calculate the signature for the data, given a key.

        :param key: Key for the signature
        :type key: str
        :param data: Data to sign
        :type data: str
        :return: The SHA256 signature
        :rtype: bytes
        """

        signature = hmac.new(key, data.encode("utf-8"), hashlib.sha256)
        return signature.hexdigest()

    def __decrypt(self, secret_token: str, data: str) -> str:
        """Decrypt data from the server using the provided token.

        :param secret_token: Token for the server
        :type secret_token: str
        :param data: The data to decrypt
        :type data: str
        :returns: Decrypted data
        :rtype: str
        """

        init_vector = secret_token[: len(secret_token) // 2]
        key = secret_token[len(secret_token) // 2 :]
        decryptor = AES.new(key, AES.MODE_CBC, init_vector)

        decrypted_data = UNPAD(decryptor.decrypt(base64.b64decode(data)))

        return decrypted_data

    def __encrypt(self, secret_token: str, data: str) -> str:
        """Encrypt data for the server using the provided token.

        :param secret_token: Token for the server
        :type secret_token: str
        :param data: The data to decrypt
        :type data: str
        :returns: Decrypted data
        :rtype: str
        """

        data = PAD(data.encode("utf-8"))
        init_vector = secret_token[: len(secret_token) // 2]
        key = secret_token[len(secret_token) // 2 :]
        encryptor = AES.new(key, AES.MODE_CBC, init_vector)
        encrypted_data = base64.b64encode(encryptor.encrypt(data))

        return encrypted_data.decode("utf-8")

    def update_request_id(self) -> None:
        """Update ``__request_id``.

        This has to be done for every new request.
        """

        self.__request_id = int(time.time())

    def connect(self, email: str, password: str) -> bool:
        """Establish a connection to the MyJD Api.

        :param email: MyJDownloader user email
        :type email: str
        :param password: MyJDownloader user email
        :type password: str
        :returns: True if successful, False if there was any error.
        :rtype: bool
        """

        self.update_request_id()
        self.__login_secret = None
        self.__device_secret = None
        self.__session_token = None
        self.__regain_token = None
        self.__server_encryption_token = None
        self.__device_encryption_token = None
        self.__devices = None
        self.__connected = False

        self.__login_secret = self.__create_secret(email, password, "server")
        self.__device_secret = self.__create_secret(email, password, "device")
        response = self.request_api(
            "/my/connect", "GET", [("email", email), ("appkey", self.__app_key)]
        )
        self.__connected = True
        self.update_request_id()
        self.__session_token = response["sessiontoken"]
        self.__regain_token = response["regaintoken"]
        self.__update_encryption_tokens()
        self.update_devices()

        return response

    def reconnect(self) -> bool:
        """Re-establish connection to the API.

        :returns: True if successful, False if there was any error.
        :rtype: bool
        """

        response = self.request_api(
            "/my/reconnect",
            "GET",
            [
                ("sessiontoken", self.__session_token),
                ("regaintoken", self.__regain_token),
            ],
        )
        self.update_request_id()
        self.__session_token = response["sessiontoken"]
        self.__regain_token = response["regaintoken"]
        self.__update_encryption_tokens()

        return response

    def disconnect(self) -> bool:
        """Disconnect from the API.

        :returns: True if successful, False if there was any error.
        :rtype: bool
        """

        response = self.request_api(
            "/my/disconnect", "GET", [("sessiontoken", self.__session_token)]
        )
        self.update_request_id()
        self.__login_secret = None
        self.__device_secret = None
        self.__session_token = None
        self.__regain_token = None
        self.__server_encryption_token = None
        self.__device_encryption_token = None
        self.__devices = None
        self.__connected = False

        return response

    def update_devices(self) -> bool:
        """Update available devices.

        Use ``list_devices()`` to get the device list.

        :returns: True if successful, False if there was any error
        :rtype: bool
        """

        response = self.request_api(
            "/my/listdevices", "GET", [("sessiontoken", self.__session_token)]
        )
        self.update_request_id()
        self.__devices = response["list"]

        return response

    def list_devices(self) -> List[Dict]:
        """Get available devices.

        Use ``update_devices()`` to update the device list.
        Each device in the list is a dictionary, like in this example:

        .. code-block :: json

            [
                {
                    "name": "Device",
                    "id": "af9d03a21ddb917492dc1af8a6427f11",
                    "type": "jd"
                }
            ]

        :returns: List of devices
        :rtype: list[dict]
        """

        return self.__devices

    def get_device(self, device_name: str = None, device_id: str = None) -> JDDevice:
        """Get a JDDevice instance for a device

        Will search for ``device_id`` first and then for ``device_name``.

        :param device_name: Name of the device
        :type device_name: str
        :param device_id: ID of the device
        :type device_id: str
        :return: JDDevice instance of the device
        :rtype: JDDevice
        """

        if not self.is_connected():
            raise (Exception("No connection established\n"))

        if device_id is not None:
            for device in self.__devices:
                if device["id"] == device_id:
                    return JDDevice(self, MyJDConnectionHelper, device)

        elif device_name is not None:
            for device in self.__devices:
                if device["name"] == device_name:
                    return JDDevice(self, MyJDConnectionHelper, device)

        raise (Exception("Device not found\n"))

    def request_api(
        self,
        path: str,
        http_method: str = "GET",
        params: Any = None,
        action: str = None,
        api: str = None,
        binary: bool = False,
    ) -> Any:
        """Make a request to the MyJD API.

        The request goes to ``path``, using the ``http_method`` with the
        parameters ``params``.

        All requests should be executed with this method.

        :param path: The path to the endpoint (excluding the base_url)
        :type path: str
        :param http_method: HTTP method for the request.
            Can be:
            - GET
            - POST
        :type http_method: str
        :param params: Parameters for the request
        :type params: dict
        :param action: On which device to execute this.
            The string could look like this: ``/t_<session_token>_<device_id>``
        :type action: str
        :param api: The API URL. (Will be set to ``self.__api_url`` by
            default.)
        :type api: str
        :returns: The response
        :rtype: Any
        """

        if not api:
            api = self.__api_url

        data = None
        if not self.is_connected() and path != "/my/connect":
            raise (Exception("No connection established\n"))

        if http_method == "GET":
            query = [path + "?"]
            if params is not None:
                for param in params:
                    if param[0] != "encryptedLoginSecret":
                        query += ["%s=%s" % (param[0], quote(param[1]))]
                    else:
                        query += ["&%s=%s" % (param[0], param[1])]
            query += ["rid=" + str(self.__request_id)]

            if self.__server_encryption_token is None:
                query += [
                    "signature="
                    + str(
                        self.__create_signature(
                            self.__login_secret, query[0] + "&".join(query[1:])
                        )
                    )
                ]

            else:
                query += [
                    "signature="
                    + str(
                        self.__create_signature(
                            self.__server_encryption_token,
                            query[0] + "&".join(query[1:]),
                        )
                    )
                ]

            query = query[0] + "&".join(query[1:])
            encrypted_response = requests.get(api + query, timeout=3)
        else:
            params_request = []
            if params is not None:
                for param in params:
                    if not isinstance(param, list):
                        params_request += [json.dumps(param)]
                    else:
                        params_request += [param]

            params_request = {
                "apiVer": self.__api_version,
                "url": path,
                "params": params_request,
                "rid": self.__request_id,
            }

            data = json.dumps(params_request)
            # Removing quotes around null elements.
            data = data.replace('"null"', "null")
            data = data.replace("'null'", "null")
            encrypted_data = self.__encrypt(self.__device_encryption_token, data)

            if action is not None:
                request_url = api + action + path
            else:
                request_url = api + path

            try:
                encrypted_response = requests.post(
                    request_url,
                    headers={"Content-Type": "application/aesjson-jd; charset=utf-8"},
                    data=encrypted_data,
                    timeout=3,
                )

            except requests.exceptions.RequestException:
                return None

        if encrypted_response.status_code != 200:
            try:
                error_msg = json.loads(encrypted_response.text)
            except json.JSONDecodeError:
                try:
                    error_msg = json.loads(
                        self.__decrypt(
                            self.__device_encryption_token, encrypted_response.text
                        )
                    )
                except json.JSONDecodeError:
                    raise Exception(
                        "Failed to decode response: {}", encrypted_response.text
                    )

            msg = (
                "\n\tSOURCE: "
                + error_msg["src"]
                + "\n\tTYPE: "
                + error_msg["type"]
                + "\n------\nREQUEST_URL: "
                + api
                + path
            )

            if http_method == "GET":
                msg += query

            msg += "\n"
            if data is not None:
                msg += "DATA:\n" + data

            raise (Exception(msg))

        if binary:

            self.update_request_id()

            # Binary content is not encrypted
            return encrypted_response.content

        if action is None:
            if not self.__server_encryption_token:
                response = self.__decrypt(self.__login_secret, encrypted_response.text)
            else:
                response = self.__decrypt(
                    self.__server_encryption_token, encrypted_response.text
                )

        else:
            response = self.__decrypt(
                self.__device_encryption_token, encrypted_response.text
            )

        jsondata = json.loads(response.decode("utf-8"))
        if jsondata["rid"] != self.__request_id:
            self.update_request_id()
            return None

        self.update_request_id()
        return jsondata



from .jd_device import JDDevice
from .direct_connection_helper import DirectConnectionHelper
import requests


class DirectConnector:
    def __init__(self, base_url: str = "http://localhost:3128"):

        self.base_url = base_url

    def is_connected(self) -> bool:
        """Check if the JDownloader is reachable.

        This makes a dummy request to the JDownloader and returns True if it
        was successful

        :returns: Connection status
        :rtype: bool
        """

        try:
            requests.get(self.base_url + "/jd/version")
            return True
        except Exception:
            pass

        return False

    def get_device(self):

        device_dict = {
            "id": "local",
            "name": "Local JDownloader",
            "type": "jd",
        }
        return JDDevice(self, DirectConnectionHelper, device_dict)



from .jd_types import CaptchaJob, SkipRequest
from typing import Any, List, TYPE_CHECKING

if TYPE_CHECKING:
    from .jd_device import JDDevice


class Captcha:
    def __init__(self, device: "JDDevice") -> None:

        self.device = device
        self.endpoint = "captcha"

    def action(self, route: str, params: Any = None, binary: bool = False) -> Any:
        route = f"/{self.endpoint}{route}"
        return self.device.connection_helper.action(route, params, binary=binary)

    def get(self, captcha_id: int, c_format: str = None) -> str:
        """Get the base64 captcha image.

        The result is a captcha image as base64 encoded data url.

        :param captcha_id: The ID of the captcha.
        :type captcha_id: int
        :param format: The format
        :type format: str
        :return: Captcha image (base64 encoded).
        :rtype: str
        """

        params = [str(captcha_id)]
        if c_format:
            params.append(c_format)
        resp = self.action("/get", params, binary=True)
        return resp

    def get_captcha_job(self, job_id: int) -> CaptchaJob:
        """Get a captcha job for `job_id`

        :param job_id: ID of the job
        :type job_id: int
        :return: The captcha job object
        :rtype: CaptchaJob
        """

        params = [job_id]
        resp = self.action("/getCaptchaJob", params)
        captcha_job = CaptchaJob(**resp)
        return captcha_job

    def list(self) -> List[CaptchaJob]:
        """Get the waiting captchas

        :return: Returns a list of all available captcha jobs
        :rtype: List[CaptchaJob]
        """

        resp = self.action("/list", None)
        captcha_jobs = []
        for job in resp:
            captcha_job = CaptchaJob(**job)
            captcha_jobs.append(captcha_job)

        return captcha_jobs

    def skip(
        self, captcha_id: int, skip_type: SkipRequest = SkipRequest.SINGLE
    ) -> bool:
        """Skip a captcha with a SkipRequest type

        :param captcha_id: ID of the captcha to skip
        :type captcha_id: int
        :param skip_type: The SkipRequest type
        :type skip_type: jd_types.SkipRequest
        :return: Success
        :rtype: boolean
        """

        params = [captcha_id, skip_type.value]
        resp = self.action("/skip", params)
        return resp

    def solve(self, captcha_id: int, result: str, result_format: str = None) -> bool:
        """Solve a captcha.

        :param captcha_id: The ID of the captcha that is solved.
        :type captcha_id: int
        :param result: The solution of the captcha.
        :type result: str
        :param result_format: Format of the result
        :type result_format: str
        :return: Success
        :rtype: boolean
        """

        params = [captcha_id, result]
        if result_format:
            params.append(result_format)
        resp = self.action("/solve", params)
        return resp



from .accounts import Accounts
from .captcha import Captcha
from .config import Config
from .content import Content
from .device import Device
from .downloads import Downloads
from .events import Events
from .extensions import Extensions
from .linkgrabber import LinkGrabber
from .log import Log
from .plugins import Plugins
from .polling import Polling
from .system import System
from .toolbar import Toolbar
from .ui import UI
from .update import Update
from typing import Any


class JDDevice:
    """A class that represents a JDownloader device and its functions."""

    def __init__(self, connector: Any, connection_helper: Any, device_dict: dict):
        """Initializes the device instance.

        :param connector: The connector object (direct or MyJD)
        :type connector: Any
        :param device_dict: Dictionary with device properties
        :type device_dict: dict
        :returns: A JDDevice object
        :rtype: JDDevice
        """

        self.name = device_dict["name"]
        self.device_id = device_dict["id"]
        self.device_type = device_dict["type"]

        self.connector = connector
        self.connection_helper = connection_helper(self)

        self.accounts = Accounts(self)
        self.captcha = Captcha(self)
        self.config = Config(self)
        self.content = Content(self)
        self.device = Device(self)
        self.downloads = Downloads(self)
        self.events = Events(self)
        self.extensions = Extensions(self)
        self.linkgrabber = LinkGrabber(self)
        self.log = Log(self)
        self.plugins = Plugins(self)
        self.polling = Polling(self)
        self.system = System(self)
        self.toolbar = Toolbar(self)
        self.ui = UI(self)
        self.update = Update(self)



import time
from typing import List, TYPE_CHECKING

if TYPE_CHECKING:
    from .jd_device import JDDevice


class MyJDConnectionHelper:
    def __init__(self, device: "JDDevice") -> None:

        self.device = device

        self.__direct_connection_info = None
        self.__refresh_direct_connections()
        self.__direct_connection_enabled = True
        self.__direct_connection_cooldown = 0
        self.__direct_connection_consecutive_failures = 0

    def __refresh_direct_connections(self) -> None:
        """Check again if a direct connection is possible."""

        response = self.device.connector.request_api(
            "/device/getDirectConnectionInfos", "POST", None, self.__action_url()
        )

        if (
            response is not None
            and "data" in response
            and "infos" in response["data"]
            and len(response["data"]["infos"]) != 0
        ):

            self.__update_direct_connections(response["data"]["infos"])

    def __update_direct_connections(self, direct_info: dict) -> None:
        """Update the direct_connection info while keeping the correct order.

        :param direct_info: Information about direct connections
        :type direct_info: dict
        """

        tmp = []
        if self.__direct_connection_info is None:
            for conn in direct_info:
                tmp.append({"conn": conn, "cooldown": 0})

            self.__direct_connection_info = tmp
            return

        # Remove old connections that are not available any longer.
        for i in self.__direct_connection_info:
            if i["conn"] not in direct_info:
                tmp.remove(i)
            else:
                direct_info.remove(i["conn"])

        # Add new connections
        for conn in direct_info:
            tmp.append({"conn": conn, "cooldown": 0})

        self.__direct_connection_info = tmp

    def enable_direct_connection(self) -> None:
        """Enable direct connections."""

        self.__direct_connection_enabled = True
        self.__refresh_direct_connections()

    def disable_direct_connect(self) -> None:
        """Disable direct connections."""

        self.__direct_connection_enabled = False
        self.__direct_connection_info = None

    def action(
        self,
        path: str,
        params: List = [],
        http_action: str = "POST",
        binary: bool = False,
    ) -> None:
        """Execute any action for the device using the params.

        All the information about the parameters and their default values,
        types, etc. can be found in the API specification for MyJDownloader
        (https://my.jdownloader.org/developers/).

        :param path: The URL of the endpoint (excluding the base url)
        :type path: str
        :param params: URL parameters, in a list of tuples.
            Example: ``[("param1","ex"),("param2","ex2")]`` becomes
            ``/example?param1=ex&param2=ex2``
        :type params: List
        :param http_action: The HTTP request type ('GET' or 'POST')
        :type http_action: str
        :param binary: Return binary response, if needed
        :type binary: bool
        :return: Response from the MyJD API
        :rtype: dict
        """

        action_url = self.__action_url()

        if (
            not self.__direct_connection_enabled
            or self.__direct_connection_info is None
            or time.time() < self.__direct_connection_cooldown
        ):

            # No direct connection available, use the MyJD API
            response = self.device.connector.request_api(
                path, http_action, params, action_url, binary=binary
            )

            if response is None:
                return False
            else:
                if (
                    self.__direct_connection_enabled
                    and time.time() >= self.__direct_connection_cooldown
                ):
                    self.__refresh_direct_connections()

                return response["data"]

        else:

            # A direct connection is available, try to use it
            for conn in self.__direct_connection_info:

                if time.time() > conn["cooldown"]:

                    # Use the direct connection
                    connection = conn["conn"]
                    api = f'http://{connection["ip"]}:{connection["port"]}'

                    response = self.device.connector.request_api(
                        path, http_action, params, action_url, api, binary=binary
                    )

                    if response is not None:

                        # This connection worked, push it to the top of the
                        # list.
                        self.__direct_connection_info.remove(conn)
                        self.__direct_connection_info.insert(0, conn)
                        self.__direct_connection_consecutive_failures = 0

                        return response["data"]

                    else:

                        # Don't try this connection for a minute.
                        conn["cooldown"] = time.time() + 60

            # None of the direct connections worked, set a cooldown for all
            # direct connections
            self.__direct_connection_consecutive_failures += 1
            self.__direct_connection_cooldown = time.time() + (
                60 * self.__direct_connection_consecutive_failures
            )

            # Use the MyJD API instead
            response = self.device.connector.request_api(
                path, http_action, params, action_url, binary=binary
            )

            if response is None:
                return False

            self.__refresh_direct_connections()

            return response["data"]

    def __action_url(self) -> str:
        """Generate the action url for the device and session."""

        return (
            "/t_"
            + self.device.connector.get_session_token()
            + "_"
            + self.device.device_id
        )



from .jd_types import Account, AccountQuery, BasicAuth, BasicAuthType
from typing import Any, Dict, List, TYPE_CHECKING

if TYPE_CHECKING:
    from .jd_device import JDDevice


class Accounts:
    def __init__(self, device: "JDDevice"):

        self.device = device
        self.endpoint = "accountsV2"

    def action(self, route: str, params: Any = None) -> Any:
        route = f"/{self.endpoint}{route}"
        return self.device.connection_helper.action(route, params)

    def add_account(self, premium_hoster: str, username: str, password: str) -> None:
        """Add a premium hoster account.

        .. warning:: The password is used in plain text, so beware...

        :param premium_hoster: The name of the premium hoster.
            See ``list_premium_hoster`` for possible names.
        :type premium_hoster: str
        :param username: User name or email address for that account.
        :type username: str
        :param password: Password for that account.
        :type password: str
        :return: empty
        :rtype: None
        """

        params = [premium_hoster, username, password]
        self.action("/addAccount", params)

    def add_basic_auth(
        self, auth_type: BasicAuthType, hostmask: str, username: str, password: str
    ) -> int:
        """Add a basic auth account.

        These accounts are plain HTTP or FTP connections, with passwords.

        :param auth_type: The auth type has to be on of the following:

            - "HTTP"
            - "FTP"
        :type auth_type: BasicAuthType
        :param hostmask: Mask of the account host.
        :type hostmask: str
        :param username: User name for the account.
        :type username: str
        :param password: Password for the account.
        :type password: str
        :return: The ID of the newly created basic auth account.
        :rtype: int
        """

        params = [auth_type.value, hostmask, username, password]
        resp = self.action("/addBasicAuth", params)
        return resp

    def disable_accounts(self, account_ids: List[int]) -> None:
        """Disable premium hoster accounts.

        :param account_ids: A list of account uuids.
        :type account_ids: List[int]
        :return: empty
        :rtype: None
        """

        params = [account_ids]
        self.action("/disableAccounts", params)

    def enable_accounts(self, account_ids: List[int]) -> None:
        """Enable premium hoster accounts.

        :param account_ids: A list of account uuids.
        :type account_ids: List[int]
        """

        params = [account_ids]
        self.action("/enableAccounts", params)

    def get_premium_hoster_url(self, hoster: str) -> str:
        """Get the url for a premium hoster.

        Note: The url will be a redirect over the jdownloader.org servers.

        :param hoster: The name of a hoster.
        :type hoster: str
        :return: The url for ``hoster``.
        :rtype: str
        """

        params = [hoster]
        resp = self.action("/getPremiumHosterUrl", params)
        return resp

    def list_accounts(
        self, account_query: AccountQuery = AccountQuery.default()
    ) -> List[Account]:
        """List premium hoster accounts.

        :param params: An AccountQuery object.
        :type: AccountQuery
        :return: A list of accounts
        :rtype: List[Account]
        """

        params = [account_query.dict()]
        resp = self.action("/listAccounts", params)

        accounts = []
        for acc in resp:
            account = Account(**acc)
            accounts.append(account)

        return accounts

    def list_basic_auth(self) -> List[BasicAuth]:
        """List basic auth accounts.

        :return: A list of basic auth accounts.
        :rtype: list[BasicAuthentication]
        """

        resp = self.action("/listBasicAuth")

        basic_auths = []
        for auth in resp:
            basic_auth = BasicAuth(**auth)
            basic_auths.append(basic_auth)

        return basic_auths

    def list_premium_hoster(self) -> List[str]:
        """List known premium hosters.

        :return: A list of all known premium hosters.
        :rtype: List[str]
        """

        resp = self.action("/listPremiumHoster")
        return resp

    def list_premium_hoster_urls(self) -> Dict[str, str]:
        """List known premium hosters with urls.

        :return: A map of all known premium hosters to urls.
        :rtype: Dict[str, str]
        """

        resp = self.action("/listPremiumHosterUrls")
        return resp

    def refresh_accounts(self, account_ids: List[int]) -> None:
        """Let JDownloader refresh the account status for ``account_ids``.

        :param account_ids: A list of account ids
        :type account_ids: List[int]
        :return: empty
        :rtype: None
        """

        params = [account_ids]
        self.action("/refreshAccounts", params)

    def remove_accounts(self, account_ids: List[int]) -> None:
        """Remove the accounts for ``account_ids``.

        :param account_ids: A list of account ids
        :type account_ids: List[int]
        :return: empty
        :rtype: None
        """

        params = [account_ids]
        self.action("/removeAccounts", params)

    def remove_basic_auths(self, basic_auth_ids: List[int]) -> bool:
        """Remove basic auths for ``basic_auth_ids``.

        :param basic_auth_ids: A list of basic auth ids.
        :type account_ids: List[int]
        :return: Success.
        :rtype: bool
        """

        params = [basic_auth_ids]
        resp = self.action("/removeBasicAuths", params)
        return resp

    def set_username_and_password(
        self, account_id: int, username: str, password: str
    ) -> bool:
        """Set a new username and password for a premium hoster account.

        :param account_id: The ID of the account.
        :type account_id: int
        :param username: The new username.
        :type username: str
        :param password: The new password.
        :type password: str
        :return: Success
        :rtype: bool
        """

        params = [account_id, username, password]
        resp = self.action("/setUserNameAndPassword", params)
        return resp

    def update_basic_auth(self, basic_authentication: BasicAuth) -> bool:
        """Update the credentials for a basic auth.

        .. note ::

            It is recommended to use the result of a ``list_basic_auth`` query
            for the input parameters. This way all necessary fields are filled
            and you can just update select parameters.

        :param basic_authentication: The new account parameters.
        :type basic_authentication: dict, BasicAuthentication
        :return: Success
        :rtype: bool
        """

        params = [basic_authentication.dict()]
        resp = self.action("/updateBasicAuth", params)
        return resp



from .jd_types import AdvancedConfigAPIEntry, AdvancedConfigQuery, EnumOption
from typing import Any, List, TYPE_CHECKING

if TYPE_CHECKING:
    from .jd_device import JDDevice


class Config:
    def __init__(self, device: "JDDevice") -> None:
        self.device = device
        self.endpoint = "config"

    def action(self, route: str, params: Any = None) -> Any:
        route = f"/{self.endpoint}{route}"
        return self.device.connection_helper.action(route, params)

    def get(self, interface_name: str, storage: str, key: str) -> Any:
        """Get value from interface by key.

        :param interface_name: The name of the JDownloader interface
        :type interface_name: str
        :param storage: The storage for the config entry.
            (None, for normal settings, or the extensions name for extension
            settings)
        :type storage: str
        :param key: The key of the config entry
        :type key: str
        :returns: The value of the config entry
        :rtype: Any
        """

        params = [interface_name, storage, key]
        resp = self.action("/get", params)
        return resp

    def get_default(self, interface_name: str, storage: str, key: str) -> Any:
        """Get default value from interface by key.

        :param interface_name: The name of the JDownloader interface
        :type interface_name: str
        :param storage: The storage for the config entry.
            (None, for normal settings, or the extensions name for extension
            settings)
        :type storage: str
        :param key: The key of the config entry
        :type key: str
        :returns: The default value of the config entry
        :rtype: Any
        """

        params = [interface_name, storage, key]
        resp = self.action("/getDefault", params)
        return resp

    def list(
        self,
        pattern: str = "",
        returnDescription: bool = True,
        returnValues: bool = True,
        returnDefaultValues: bool = True,
        returnEnumInfo: bool = True,
    ) -> List[AdvancedConfigAPIEntry]:
        """List all available config entries.

        :param pattern: A regex pattern to query by. If no pattern is given,
            all config items will be returned
        :type pattern: str
        :param returnDescription: If description should be returned
        :type returnDescription: boolean
        :param returnValues: If values should be returned
        :type returnValues: boolean
        :param returnDefaultValues: If default values should be returned
        :type returnDefaultValues: boolean
        :param returnEnumInfo: If enum info should be returned
        :type returnEnumInfo: boolean
        :return: A list of (matching) config items
        :rtype: List[AdvancedConfigAPIEntry]
        """

        params = [
            pattern,
            returnDescription,
            returnValues,
            returnDefaultValues,
            returnEnumInfo,
        ]
        resp = self.action("/list", params)

        config_api_entries = []
        for entry in resp:
            config_api_entry = AdvancedConfigAPIEntry(**entry)
            config_api_entries.append(config_api_entry)

        return config_api_entries

    def list_enum(self, enum_type: str) -> List[EnumOption]:
        """List all possible enum values for the type.

        The enum_type is the AdvancedConfigAPIEntry.config_type for an Enum.
        (e.g.: 'org.jdownloader.settings.DelayWriteMode')

        :param enum_type: The enum type
        :type enum_type: str
        :returns: A list of possible options for the enum_type
        :rtype: List[EnumOption]
        """

        params = [enum_type]
        resp = self.action("/listEnum", params)

        enum_options = []
        for entry in resp:
            enum_option = EnumOption(**entry)
            enum_options.append(enum_option)

        return enum_options

    def query(
        self, advanced_config_query: AdvancedConfigQuery = AdvancedConfigQuery.default()
    ) -> List[AdvancedConfigAPIEntry]:
        """Query config entries with an :class:`AdvancedConfigQuery`.

        :param advanced_config_query: The query options
        :type advanced_config_query: AdvancedConfigQuery
        :returns: A list of config entries
        :rtype: List[AdvancedConfigAPIEntry]
        """

        params = [advanced_config_query.dict()]
        resp = self.action("/query", params=params)

        config_api_entries = []
        for entry in resp:
            config_api_entry = AdvancedConfigAPIEntry(**entry)
            config_api_entries.append(config_api_entry)

        return config_api_entries

    def reset(self, interface_name: str, storage: str, key: str) -> bool:
        """Reset a config entry.

        :param interface_name: The name of the JDownloader interface
        :type interface_name: str
        :param storage: The storage for the config entry.
            (None, for normal settings, or the extensions name for extension
            settings)
        :type storage: str
        :param key: The key of the config entry
        :type key: str
        :returns: Success
        :rtype: bool
        """

        params = [interface_name, storage, key]
        resp = self.action("/reset", params)
        return resp

    def set(self, interface_name: str, storage: str, key: str, value: str) -> bool:
        """Set a config entry.

        :param interface_name: The name of the JDownloader interface
        :type interface_name: str
        :param storage: The storage for the config entry.
            (None, for normal settings, or the extensions name for extension
            settings)
        :type storage: str
        :param key: The key of the config entry
        :type key: str
        :param value: The new value for the config. Integers, booleans, etc.
            should be converted to strings.
        :type value: str
        :returns: Success
        :rtype: bool
        """

        params = [interface_name, storage, key, value]
        resp = self.action("/set", params)
        return resp



from .jd_types import APIQuery
from typing import Any


class Polling:
    def __init__(self, device):
        self.device = device
        self.endpoint = "polling"

    def action(self, route: str, params: Any = None) -> Any:
        route = f"/{self.endpoint}{route}"
        return self.device.connection_helper.action(route, params)

    def poll(self, query_params=APIQuery.default()):
        """Poll for APIQuery."""

        params = [query_params.dict()]
        resp = self.action("/poll", params)
        return resp



from .jd_types import LinkCheckResult
from typing import Any


class Toolbar:
    def __init__(self, device):
        self.device = device
        self.endpoint = "toolbar"

    def action(self, route: str, params: Any = None) -> Any:
        route = f"/{self.endpoint}{route}"
        return self.device.connection_helper.action(route, params)

    def add_links_from_dom(self):
        """Unknown."""

        resp = self.action("/addLinksFromDOM")
        return resp

    def check_links_from_dom(self):
        """Unknown."""

        resp = self.action("/checkLinksFromDOM")
        return resp

    def get_status(self):
        """Get JDownloader status.

        Example result:

        .. code-block :: json

            {
                "running" : false,
                "reconnect" : false,
                "premium" : true,
                "download_complete" : 0,
                "stopafter" : false,
                "limit" : false,
                "limitspeed" : 0,
                "state" : "STOPPED_STATE",
                "download_current" : 0,
                "clipboard" : true,
                "speed" : 0,
                "pause" : false
            }

        """

        resp = self.action("/getStatus")
        return resp

    def is_available(self):
        """Unknown."""

        resp = self.action("/isAvailable")
        return resp

    def poll_checked_links_from_dom(self, check_id):
        """Unknown."""

        params = [check_id]
        resp = self.action("/pollCheckedLinksFromDOM", params)
        link_check_result = LinkCheckResult(**resp)
        return link_check_result

    def special_url_handling(self, url):
        """Unknown."""

        params = [url]
        resp = self.action("/specialURLHandling", params)
        return resp

    def start_downloads(self):
        """Start downloads."""

        resp = self.action("/startDownloads")
        return resp

    def stop_downloads(self):
        """Stops the downloads."""

        resp = self.action("/stopDownloads")
        return resp

    def toggle_automatic_reconnect(self):
        """Unknown."""

        resp = self.action("/toggleAutomaticReconnect")
        return resp

    def toggle_clipboard_monitoring(self):
        """Unknown."""

        resp = self.action("/toggleClipboardMonitoring")
        return resp

    def toggle_download_speed_limit(self):
        """Unknown."""

        resp = self.action("/toggleDownloadSpeedLimit")
        return resp

    def toggle_pause_downloads(self):
        """Unknown."""

        resp = self.action("/togglePauseDownloads")
        return resp

    def toggle_premium(self):
        """Unknown."""

        resp = self.action("/togglePremium")
        return resp

    def toggle_stop_after_current_download(self):
        """Unknown."""

        resp = self.action("/toggleStopAfterCurrentDownload")
        return resp

    def trigger_update(self):
        """Unknown."""

        resp = self.action("/triggerUpdate")
        return resp



from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    from .jd_device import JDDevice


class Update:
    def __init__(self, device: "JDDevice") -> None:
        self.device = device
        self.endpoint = "update"

    def action(self, route: str, params: Any = None) -> Any:
        route = f"/{self.endpoint}{route}"
        return self.device.connection_helper.action(route, params)

    def is_update_available(self) -> str:
        """Returns if an update is available."""

        resp = self.action("/isUpdateAvailable")
        return resp

    def restart_and_update(self) -> str:
        """Restarts and update."""

        resp = self.action("/restartAndUpdate")
        return resp

    def run_update_check(self) -> str:
        """Runs an update check."""

        resp = self.action("/runUpdateCheck")
        return resp



from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    from .jd_device import JDDevice


class System:
    def __init__(self, device: "JDDevice") -> None:
        self.device = device
        self.endpoint = "system"

    def action(self, route: str, params: Any = None) -> Any:
        route = f"/{self.endpoint}{route}"
        return self.device.connection_helper.action(route, params)

    def exit_jd(self) -> str:
        """Stop the JDownloader."""

        resp = self.action("/exitJD")
        return resp

    def get_storage_infos(self, path: str = None) -> dict:
        """Get storage information."""

        params = [path]
        resp = self.action("/getStorageInfos", params)
        return resp

    def get_system_infos(self) -> dict:
        """Get system information."""

        resp = self.action("/getSystemInfos")
        return resp

    def hibernate_os(self) -> str:
        """Hibernate the OS."""

        resp = self.action("/hibernateOS")
        return resp

    def restart_jd(self) -> str:
        """Restart the JDownloader."""

        resp = self.action("/restartJD")
        return resp

    def shutdown_os(self, force: bool = False) -> str:
        """Shutdown the OS."""

        params = [force]
        resp = self.action("/shutdownOS", params)
        return resp

    def standby_os(self) -> str:
        """Put the OS in standby."""

        resp = self.action("/standbyOS")
        return resp



from typing import Any
import json
import requests


class DirectConnectionHelper:
    def __init__(self, device):

        self.device = device

    def action(
        self,
        path: str,
        params: Any = None,
        http_action: str = "POST",
        binary: bool = False,
    ) -> Any:
        """Make the request to the JDownloader.

        :param path: The URL endpoint (excluding base_url) that is called.
        :type path: str
        :param params: Parameters for the request
        :type params: list, dict or str
        :param http_action: The HTTP method (unused)
        :type http_action: str
        :param binary: Return the response as byte array
        :type binary: bool
        :returns: The result of the request
        :rtype: byte_array, dict, string
        """

        rurl = f"{self.device.connector.base_url}{path}"

        param_list = []
        if params:
            for param in params:
                param_list.append(json.dumps(param))
        rparams = "?" + "&".join(param_list)

        if binary:
            return requests.get(rurl + rparams).content

        rstr = requests.get(rurl + rparams).content.decode()
        robj = json.loads(rstr)

        if "data" in robj:
            robj = robj["data"]

        return robj



from .jd_types import (
    DeleteAction,
    DownloadLink,
    FilePackage,
    LinkQuery,
    Mode,
    PackageQuery,
    Priority,
    Reason,
    SelectionType,
)
from typing import Any, Dict, List, Optional


class Downloads:
    def __init__(self, device):
        self.device = device
        self.endpoint = "downloadsV2"

    def action(self, route: str, params: Any = None) -> Any:
        route = f"/{self.endpoint}{route}"
        return self.device.connection_helper.action(route, params)

    def cleanup(
        self,
        link_ids: List[int],
        package_ids: List[int],
        delete_action: DeleteAction,
        mode: Mode,
        selection_type: SelectionType,
    ) -> Any:
        """Cleanup the link_ids & package_ids in the download list.

        :param link_ids: Link IDs that are used
        :type link_ids: List[int]
        :param package_ids: Package IDs that are used
        :type package_ids: List[int]
        :param action: The class:`jd_types.DeleteAction` that will be performed
        :type action: DeleteAction
        :param mode: The class:`jd_types.Mode` that is used
        :type mode: Mode
        :type selection_type: The class:`jd_types.SelectionType` that is
            applied
        :type selection_type: SelectionType
        :returns: resp
        :rtype: Any
        """

        params = [
            link_ids,
            package_ids,
            delete_action.value,
            mode.value,
            selection_type.value,
        ]
        resp = self.action("/cleanup", params)
        return resp

    def force_download(self, link_ids: List[int], package_ids: List[int]) -> bool:
        """Force downloads for link_ids and package_ids.

        :param link_ids: Link IDs that are used
        :type link_ids: List[int]
        :param package_ids: Package IDs that are used
        :type package_ids: List[int]
        :returns: Success
        :rtype: bool
        """

        params = [link_ids, package_ids]
        resp = self.action("/forceDownload", params)
        return resp

    def get_download_urls(
        self, link_ids: List[int], package_ids: List[int], url_display_type: List[str]
    ) -> Dict[str, List[int]]:
        """Force downloads for link_ids and package_ids.

        :param link_ids: Link IDs that are used
        :type link_ids: List[int]
        :param package_ids: Package IDs that are used
        :type package_ids: List[int]
        :param url_display_type: The type of urls that should be returned.
            Example: ['ORIGIN']
        :type url_display_type: List[str]
        :returns: The download urls with their associated packages
        :rtype: Dict[str, List[int]]
        """

        params = [link_ids, package_ids, url_display_type]
        resp = self.action("/getDownloadUrls", params)
        return resp

    def get_stop_mark(self) -> int:
        """Get the link id for where the stop mark is at.

        If no stop mark is set, the result it -1

        :returns: Link id for stop mark, or -1
        :rtype: int
        """

        resp = self.action("/getStopMark")
        return resp

    def get_stop_marked_link(self) -> Optional[DownloadLink]:
        """Get the :class:`DownloadLink` object for the stopmark.

        :returns: Download link for stop mark, or None
        :rtype: DownloadLink
        """

        resp = self.action("/getStopMarkedLink")

        if resp:
            download_link = DownloadLink(**resp)
            return download_link

        return None

    def get_structure_change_counter(self, old_counter_value: int) -> int:
        """Get the structure change counter.

        Update the application layout, if the structure_change_counter is
        higher than the last.

        :returns: Structure change counter, or -1 if there is no newer change.
        :rtype: int
        """

        params = [old_counter_value]
        resp = self.action("/getStructureChangeCounter", params)
        return resp

    def move_links(
        self, link_ids: List[int], after_link_id: int, dest_package_id: int
    ) -> Any:
        """Move links to a package.

        :param link_ids: Link IDs that are used
        :type link_ids: List[int]
        :param after_link_id: ?
        :type after_link_id: int
        :param dest_package_id: The ID of the destination package
        :type dest_package_id: int
        :returns: resp
        :rtype: Any
        """

        params = [link_ids, after_link_id, dest_package_id]
        resp = self.action("/moveLinks", params)
        return resp

    def move_packages(self, package_ids: List[int], after_dest_package_id: int) -> Any:
        """Move packages.

        :param package_ids: Package IDs that are used
        :type package_ids: List[int]
        :param after_dest_package_id: ?
        :type after_dest_package_id: int
        :returns: resp
        :rtype: Any
        """

        params = [package_ids, after_dest_package_id]
        resp = self.action("/movePackages", params)
        return resp

    def move_to_new_package(
        self,
        link_ids: List[int],
        pkg_ids: List[int],
        new_pkg_name: str,
        download_path: str,
    ) -> Any:
        """Move link_ids and pkg_ids to a new package

        :param link_ids: Link IDs that are used
        :type link_ids: List[int]
        :param package_ids: Package IDs that are used
        :type package_ids: List[int]
        :param new_pkg_name: Name of the new package
        :type new_pkg_name: str
        :param download_path: Download path for the new package
        :type download_path: str
        :returns: resp
        :rtype: Any
        """

        params = [link_ids, pkg_ids, new_pkg_name, download_path]
        resp = self.action("/movetoNewPackage", params)
        return resp

    def package_count(self) -> int:
        """Get the number of packages in the download list.

        :returns: Number of packages in download list
        :rtype: int
        """

        resp = self.action("/packageCount")
        return resp

    def query_links(
        self, query_params: LinkQuery = LinkQuery.default()
    ) -> List[DownloadLink]:
        """Query the links in the download list.

        :param query_params: The parameters for the query
        :type query_params: LinkQuery
        :returns: A list of download link objects
        :rtype: List[DownloadLink]
        """

        params = [query_params.dict()]
        resp = self.action("/queryLinks", params)

        download_links = []
        for link in resp:
            download_link = DownloadLink(**link)
            download_links.append(download_link)

        return download_links

    def query_packages(
        self, query_params: PackageQuery = PackageQuery.default()
    ) -> List[FilePackage]:
        """Query the packages in the download list.

        :param query_params: The parameters for the query
        :type query_params: PackageQuery
        :returns: A list of file packages objects
        :rtype: List[FilePackage]
        """

        params = [query_params.dict()]
        resp = self.action("/queryPackages", params)

        download_packages = []
        for package in resp:
            download_package = FilePackage(**package)
            download_packages.append(download_package)

        return download_packages

    def remove_links(self, link_ids: List[int], package_ids: List[int]) -> None:
        """Remove links/packages from download list.

        :param link_ids: Link IDs that are used
        :type link_ids: List[int]
        :param package_ids: Package IDs that are used
        :type package_ids: List[int]
        """

        params = [link_ids, package_ids]
        self.action("/removeLinks", params)

    def remove_stop_mark(self) -> None:
        """Remove the stop mark."""

        self.action("/removeStopMark")

    def rename_link(self, link: int, new_name: str) -> None:
        """Rename a link.

        :param link: The ID of the link
        :type link: int
        :param new_name: The new name for the link
        :type new_name: str
        """

        params = [link, new_name]
        self.action("/renameLink", params)

    def rename_package(self, package_id: str, new_name: str) -> None:
        """Rename a package.

        :param package_id: ID of the packages
        :type package_id: int
        :param new_name: New name for the package
        :type new_name: str
        """

        params = [package_id, new_name]
        resp = self.action("/renamePackage", params)
        return resp

    def reset_links(self, link_ids: List[int], package_ids: List[int]) -> None:
        """Reset links/packages in the download list.

        :param link_ids: Link IDs that are used
        :type link_ids: List[int]
        :param package_ids: Package IDs that are used
        :type package_ids: List[int]
        """

        params = [link_ids, package_ids]
        self.action("/resetLinks", params)

    def resume_links(self, link_ids: List[int], package_ids: List[int]) -> None:
        """Resume links/packages.

        :param link_ids: Link IDs that are used
        :type link_ids: List[int]
        :param package_ids: Package IDs that are used
        :type package_ids: List[int]
        """

        params = [link_ids, package_ids]
        self.action("/resumeLinks", params)

    def set_download_directory(self, directory: str, package_ids: List[int]) -> None:
        """Set the download directory for a packages.

        :param directory: Path of the download directory
        :type directory: str
        :param package_ids: List of package IDs that are changed.
        :type package_ids: List[int]
        """

        params = [directory, package_ids]
        resp = self.action("/setDownloadDirectory", params)
        return resp

    def set_download_password(
        self, link_ids: List[int], package_ids: List[int], password: str
    ) -> bool:
        """Set the download password for links/packages.

        :param link_ids: Link IDs that are used
        :type link_ids: List[int]
        :param package_ids: List of package IDs that are changed.
        :type package_ids: List[int]
        :param password: The download password
        :type password: str
        :returns: Success
        :rtype: bool
        """

        params = [link_ids, package_ids, password]
        resp = self.action("/setDownloadPassword", params)
        return resp

    def set_enabled(
        self, enabled: bool, link_ids: List[int], package_ids: List[int]
    ) -> bool:
        """Enable/disable links and packages.

        :param enabled: Enable on or off
        :type enabled: bool
        :param link_ids: Link IDs that are used
        :type link_ids: List[int]
        :param package_ids: List of package IDs that are changed.
        :type package_ids: List[int]
        """

        params = [enabled, link_ids, package_ids]
        resp = self.action("/setEnabled", params)
        return resp

    def set_priority(
        self, priority: Priority, link_ids: List[int], package_ids: List[int]
    ) -> None:
        """Set the priority for links and packages.

        :param priority: The priority for the links/packages.
        :type priority: Priority
        :param link_ids: Link IDs that are used
        :type link_ids: List[int]
        :param package_ids: List of package IDs that are changed.
        :type package_ids: List[int]
        """

        params = [priority.value, link_ids, package_ids]
        self.action("/setPriority", params)

    def set_stop_mark(self, link_id: int = None, package_id: int = None) -> None:
        """Set the stop mark to the specified id.

        Only one of link_id and package_id has to be given.

        :param link_id: A link id for the stop mark
        :type link_id: int
        :param package_id: A package id for the stop mark
        :type package_id: int
        """

        params = [link_id, package_id]
        self.action("/setStopMark", params)

    def split_package_by_hoster(
        self, link_ids: List[int], package_ids: List[int]
    ) -> None:
        """Split the packages/links by hoster.

        :param link_ids: Link IDs that are used
        :type link_ids: List[int]
        :param package_ids: List of package IDs that are changed.
        :type package_ids: List[int]
        """

        params = [link_ids, package_ids]
        self.action("/splitPackageByHoster", params)

    def start_online_status_check(
        self, link_ids: List[int], package_ids: List[int]
    ) -> None:
        """Start an online status check for links and packages.

        :param link_ids: Link IDs that are used
        :type link_ids: List[int]
        :param package_ids: List of package IDs that are changed.
        :type package_ids: List[int]
        """

        params = [link_ids, package_ids]
        self.action("/startOnlineStatusCheck", params)

    def unskip(
        self, link_ids: List[int], package_ids: List[int], filter_by_reason: Reason
    ) -> bool:
        """Un-skip links and packages

        :param package_ids: List of package IDs that are changed.
        :type package_ids: List[int]
        :param link_ids: Link IDs that are used
        :type link_ids: List[int]
        :param filter_by_reason: Filter for the reason why they were skipped.
        :type filter_by_reason: Reason
        :returns: Success
        :rtype: bool
        """

        # package_ids and link_ids are switch for whatever reason...
        params = [package_ids, link_ids, filter_by_reason.value]
        resp = self.action("/unskip", params)
        return resp



"""
These are the types and constants that are defined in JDownloader.

For more information, see here:
    https://my.jdownloader.org/developers/index.html#tag_342
"""

from enum import Enum
from pydantic import BaseModel
from typing import Any, ForwardRef, List, Optional


#
# enums and constants
#


class AbstractType(str, Enum):
    """Abstract types that are used for config entries."""

    BOOLEAN = "BOOLEAN"
    INT = "INT"
    LONG = "LONG"
    STRING = "STRING"
    OBJECT = "OBJECT"
    OBJECT_LIST = "OBJECT_LIST"
    STRING_LIST = "STRING_LIST"
    ENUM = "ENUM"
    BYTE = "BYTE"
    CHAR = "CHAR"
    DOUBLE = "DOUBLE"
    FLOAT = "FLOAT"
    SHORT = "SHORT"
    BOOLEAN_LIST = "BOOLEAN_LIST"
    BYTE_LIST = "BYTE_LIST"
    SHORT_LIST = "SHORT_LIST"
    LONG_LIST = "LONG_LIST"
    INT_LIST = "INT_LIST"
    FLOAT_LIST = "FLOAT_LIST"
    ENUM_LIST = "ENUM_LIST"
    DOUBLE_LIST = "DOUBLE_LIST"
    CHAR_LIST = "CHAR_LIST"
    UNKNOWN = "UNKNOWN"
    HEX_COLOR = "HEX_COLOR"
    HEX_COLOR_LIST = "HEX_COLOR_LIST"
    ACTION = "ACTION"


class DeleteAction(str, Enum):
    """Delete actions, that can be executed.

    This corresponds to the "Action" enum of JDownloader.
    """

    DELETE_ALL = "DELETE_ALL"
    DELETE_DISABLED = "DELETE_DISABLED"
    DELETE_FAILED = "DELETE_FAILED"
    DELETE_FINISHED = "DELETE_FINISHED"
    DELETE_DUPE = "DELETE_DUPE"
    DELETE_MODE = "DELETE_MODE"


class AvailableLinkState(str, Enum):
    """The availability of a link."""

    ONLINE = "ONLINE"
    OFFLINE = "OFFLINE"
    UNKNOWN = "UNKNOWN"
    TEMP_UNKNOWN = "TEMP_UNKNOWN"


class BasicAuthType(str, Enum):
    """Types of basic auth protocols."""

    FTP = "FTP"
    HTTP = "HTTP"


class Context(str, Enum):
    """Contextmenu selection."""

    LGC = "LGC"  # linkgrabber rightclick
    DLC = "DLC"  # downloadlist rightclick


class MenuType(str, Enum):
    """Menu types"""

    CONTAINER = "CONTAINER"
    ACTION = "ACTION"
    LINK = "LINK"


class Mode(str, Enum):
    """Modes for package deletion."""

    REMOVE_LINKS_AND_DELETE_FILES = "REMOVE_LINKS_AND_DELETE_FILES"
    REMOVE_LINKS_AND_RECYCLE_FILES = "REMOVE_LINKS_AND_RECYCLE_FILES"
    REMOVE_LINKS_ONLY = "REMOVE_LINKS_ONLY"


class Priority(str, Enum):
    """Download priority for packages."""

    HIGHEST = "HIGHEST"
    HIGHER = "HIGHER"
    HIGH = "HIGH"
    DEFAULT = "DEFAULT"
    LOW = "LOW"
    LOWER = "LOWER"
    LOWEST = "LOWEST"


class Reason(str, Enum):
    """Reasons for exceptions."""

    CONNECTION_UNAVAILABLE = "CONNECTION_UNAVAILABLE"
    TOO_MANY_RETRIES = "TOO_MANY_RETRIES"
    CAPTCHA = "CAPTCHA"
    MANUAL = "MANUAL"
    DISK_FULL = "DISK_FULL"
    NO_ACCOUNT = "NO_ACCOUNT"
    INVALID_DESTINATION = "INVALID_DESTINATION"
    FILE_EXISTS = "FILE_EXISTS"
    UPDATE_RESTART_REQUIRED = "UPDATE_RESTART_REQUIRED"
    FFMPEG_MISSING = "FFMPEG_MISSING"
    FFPROBE_MISSING = "FFPROBE_MISSING"


class SelectionType(str, Enum):
    """Types for selection"""

    SELECTED = "SELECTED"
    UNSELECTED = "UNSELECTED"
    ALL = "ALL"
    NONE = "NONE"


class SkipRequest(str, Enum):
    """Captcha skip request"""

    SINGLE = "SINGLE"
    BLOCK_HOSTER = "BLOCK_HOSTER"
    BLOCK_ALL_CAPTCHAS = "BLOCK_ALL_CAPTCHAS"
    BLOCK_PACKAGE = "BLOCK_PACKAGE"
    REFRESH = "REFRESH"
    STOP_CURRENT_ACTION = "STOP_CURRENT_ACTION"
    TIMEOUT = "TIMEOUT"


class Status(str, Enum):
    """Status"""

    NA = "NA"
    PENDING = "PENDING"
    FINISHED = "FINISHED"


#
# structures and objects
#


class Account(BaseModel):
    """This is a premium hoster account

    Initializes itself from a query result (dict)
    """

    enabled: Optional[bool]
    errorString: Optional[str]
    errorType: Optional[str]
    hostname: Optional[str]
    trafficLeft: Optional[int]
    trafficMax: Optional[int]
    username: Optional[str]
    uuid: Optional[int]
    valid: Optional[bool]
    validUntil: Optional[int]

    def __repr__(self) -> str:
        return f"<Account ({self.uuid})>"


class AccountQuery(BaseModel):
    """Query for premium host accounts.

    The fields are booleans, that can be turned on or off, if you want to have
    query for the information or not.
    By default all possible data is queried.
    """

    enabled: bool = (True,)
    error: bool = (True,)
    maxResults: int = (-1,)
    startAt: int = (0,)
    trafficLeft: bool = (True,)
    trafficMax: bool = (True,)
    userName: bool = (True,)
    uuidlist: Optional[List[int]]
    valid: bool = (True,)
    validUntil: bool = (True,)

    def __repr__(self):
        return f"<AccountQuery ({self.uuid_list})>"

    def default():
        return AccountQuery(
            enabled=True,
            error=True,
            maxResults=-1,
            startAt=0,
            trafficLeft=True,
            trafficMax=True,
            userName=True,
            uuidlist=None,
            valid=True,
            validUntil=True,
        )


class AdvancedConfigAPIEntry(BaseModel):

    abstractType: Optional[AbstractType]
    defaultValue: Optional[Any]
    docs: Optional[str]
    enumLabel: Optional[str]
    enumOptions: Optional[Any]
    interfaceName: Optional[str]
    key: Optional[str]
    storage: Optional[str]
    type: Optional[str]
    value: Optional[Any]

    def __repr__(self):
        return f"<AdvancedConfigAPIEntry ({self.key})>"


class AdvancedConfigQuery(BaseModel):

    configInterface: Optional[str]
    defaultValues: bool
    description: bool
    enumInfo: bool
    includeExtensions: bool
    pattern: Optional[str]
    values: bool

    def __repr__(self):
        return f"<AdvancedConfigQuery ({self.config_interface})>"

    def default():
        return AdvancedConfigQuery(
            configInterface=None,
            defaultValues=True,
            description=True,
            enumInfo=True,
            includeExtensions=True,
            pattern=None,
            values=True,
        )


class BasicAuth(BaseModel):

    created: Optional[int]
    enabled: Optional[bool]
    hostmask: Optional[str]
    id: Optional[int]
    lastValidated: Optional[int]
    password: Optional[str]
    type: Optional[BasicAuthType]
    username: Optional[str]

    def __repr__(self):
        return f"<BasicAuth ({self.id})>"


class AddLinksQuery(BaseModel):

    assignJobID: Optional[bool]
    autoExtract: Optional[bool]
    autostart: Optional[bool]
    dataURLs: List[str] = []
    deepDecrypt: Optional[bool]
    destinationFolder: Optional[str]
    downloadPassword: Optional[str]
    extractPassword: Optional[str]
    links: Optional[str]
    overwritePackagizerRules: Optional[bool]
    packageName: Optional[str]
    priority: Optional[Priority] = Priority.DEFAULT
    sourceUrl: Optional[str]

    def __repr__(self):
        return f"<AddLinksQuery ({self.packageName})>"


class APIQuery(BaseModel):
    """A standard api query.

    Most endpoint use a specialized version.
    """

    empty: bool
    forNullKey: Optional[str]
    maxResults: int
    startAt: int

    def __repr__(self):
        return "<APIQuery>"

    def default():
        return APIQuery(empty=False, forNullKey="", maxResults=-1, startAt=0)


class CaptchaJob(BaseModel):

    captchaCategory: Optional[str]
    created: Optional[int]
    explain: Optional[str]
    hoster: Optional[str]
    id: Optional[int]
    link: Optional[int]
    timeout: Optional[int]
    type: Optional[str]

    def __repr__(self):
        return f"<CaptchaJob ({self.captcha_id})>"


class LinkVariant(BaseModel):

    iconKey: Optional[str]
    id: Optional[str]
    name: Optional[str]

    def __repr__(self):
        return f"<LinkVariant ({self.id})>"


class CrawledLink(BaseModel):

    availability: Optional[AvailableLinkState]
    bytesTotal: Optional[int]
    comment: Optional[str]
    downloadPassword: Optional[str]
    enabled: Optional[bool]
    host: Optional[str]
    name: Optional[str]
    packageUUID: Optional[int]
    priority: Optional[Priority]
    url: Optional[str]
    uuid: Optional[int]
    variant: Optional[LinkVariant]
    variants: Optional[bool]

    def __repr__(self):
        return f"<CrawledLink ({self.uuid})>"


class CrawledLinkQuery(BaseModel):

    availability: Optional[bool]
    bytesTotal: Optional[bool]
    comment: Optional[bool]
    enabled: Optional[bool]
    host: Optional[bool]
    jobUUIDs: Optional[List[int]]
    maxResults: Optional[int]
    packageUUIDs: Optional[List[int]]
    password: Optional[bool]
    priority: Optional[bool]
    startAt: Optional[int]
    status: Optional[bool]
    url: Optional[bool]
    variantID: Optional[bool]
    variantIcon: Optional[bool]
    variantName: Optional[bool]
    variants: Optional[bool]

    def __repr__(self):
        return "<CrawledLinkQuery>"

    def default():
        return CrawledLinkQuery(
            availability=True,
            bytesTotal=True,
            comment=True,
            enabled=True,
            host=True,
            jobUUIDs=None,
            maxResults=-1,
            packageUUIDs=None,
            password=True,
            priority=True,
            startAt=0,
            status=True,
            url=True,
            variantID=True,
            variantIcon=True,
            variantName=True,
            variants=True,
        )


class CrawledPackage(BaseModel):

    bytesTotal: Optional[int]
    childCount: Optional[int]
    comment: Optional[str]
    downloadPassword: Optional[str]
    enabled: Optional[bool]
    hosts: Optional[List[str]]
    name: Optional[str]
    offlineCount: Optional[int]
    onlineCount: Optional[int]
    priority: Optional[Priority]
    saveTo: Optional[str]
    tempUnknownCount: Optional[int]
    unknownCount: Optional[int]
    uuid: Optional[int]

    def __repr__(self):
        return f"<CrawledPackage ({self.uuid})"


class CrawledPackageQuery(BaseModel):

    availableOfflineCount: Optional[bool]
    availableOnlineCount: Optional[bool]
    availableTempUnknownCount: Optional[bool]
    availableUnknownCount: Optional[bool]
    bytesTotal: Optional[bool]
    childCount: Optional[bool]
    comment: Optional[bool]
    enabled: Optional[bool]
    hosts: Optional[bool]
    maxResults: Optional[int]
    packageUUIDs: Optional[List[int]]
    priority: Optional[bool]
    saveTo: Optional[bool]
    startAt: Optional[int]
    status: Optional[bool]

    def __repr__(self):
        return "<CrawledPackageQuery>"

    def default():
        return CrawledPackageQuery(
            availableOfflineCount=True,
            availableOnlineCount=True,
            availableTempUnknownCount=True,
            availableUnknownCount=True,
            bytesTotal=True,
            childCount=True,
            comment=True,
            enabled=True,
            hosts=True,
            maxResults=-1,
            packageUUIDs=None,
            priority=True,
            saveTo=True,
            startAt=0,
            status=True,
        )


class DownloadLink(BaseModel):

    addedDate: Optional[int]
    bytesLoaded: Optional[int]
    bytesTotal: Optional[int]
    comment: Optional[str]
    downloadPassword: Optional[str]
    enabled: Optional[bool]
    eta: Optional[int]
    extractionStatus: Optional[str]
    finished: Optional[bool]
    finishedDate: Optional[int]
    host: Optional[str]
    name: Optional[str]
    packageUUID: Optional[int]
    priority: Optional[Priority]
    running: Optional[int]
    skipped: Optional[int]
    speed: Optional[int]
    status: Optional[str]
    statusIconKey: Optional[str]
    url: Optional[str]
    uuid: Optional[int]

    def __repr__(self):
        return f"<DownloadLink ({self.uuid})>"


class EnumOption(BaseModel):

    label: Optional[str]
    name: Optional[str]

    def __repr__(self):
        return f"<EnumOption ({self.name})>"


class Extension(BaseModel):

    configInterface: Optional[str]
    description: Optional[str]
    enabled: Optional[bool]
    iconKey: Optional[str]
    id: Optional[str]
    installed: Optional[bool]
    name: Optional[str]

    def __repr__(self):
        return f"<Extension ({self.id})>"


class ExtensionQuery(BaseModel):

    configInterface: Optional[bool]
    description: Optional[bool]
    enabled: Optional[bool]
    iconKey: Optional[bool]
    installed: Optional[bool]
    name: Optional[bool]
    pattern: Optional[str]

    def __repr__(self):
        return "<ExtensionQuery>"

    def default():
        return ExtensionQuery(
            configInterface=True,
            description=True,
            enabled=True,
            iconKey=True,
            installed=True,
            name=True,
            pattern=None,
        )


class FilePackage(BaseModel):

    activeTask: Optional[str]
    bytesLoaded: Optional[int]
    bytesTotal: Optional[int]
    childCount: Optional[int]
    comment: Optional[str]
    downloadPassword: Optional[str]
    enabled: Optional[bool]
    eta: Optional[int]
    finished: Optional[bool]
    hosts: Optional[List[str]]
    name: Optional[str]
    priority: Optional[Priority]
    running: Optional[bool]
    saveTo: Optional[str]
    speed: Optional[int]
    status: Optional[str]
    statusIconKey: Optional[str]
    uuid: Optional[int]

    def __repr__(self):
        return f"<FilePackage ({self.uuid})>"


IconDescriptor = ForwardRef("IconDescriptor")


class IconDescriptor(BaseModel):

    cls: Optional[str]
    key: Optional[str]
    prps: Optional[Any]
    rsc: Optional[IconDescriptor]

    def __repr__(self):
        return f"<IconDescriptor ({self.key})>"


class JobLinkCrawler(BaseModel):

    broken: Optional[int]
    checking: Optional[bool]
    crawled: Optional[int]
    crawledId: Optional[int]
    crawling: Optional[bool]
    filtered: Optional[int]
    jobId: Optional[int]
    unhandled: Optional[int]

    def __repr__(self):
        return f"<JobLinkCrawler ({self.crawlerId})>"


class LinkStatus(BaseModel):

    host: Optional[str]
    linkCheckID: Optional[str]
    name: Optional[str]
    size: Optional[int]
    status: Optional[AvailableLinkState]
    url: Optional[str]

    def __repr__(self):
        return f"<LinkStatus ({self.link_check_id})>"


class LinkCheckResult(BaseModel):

    link: Optional[List[LinkStatus]]
    status: Optional[Status]

    def __repr__(self):
        return "<LinkCheckResult>"


class LinkCollectingJob(BaseModel):

    id: Optional[int]

    def __repr__(self):
        return f"<LinkCollectingJob ({self.id})>"


class LinkCrawlerJobsQuery(BaseModel):

    collectorInfo: Optional[bool]
    jobIds: Optional[List[int]]

    def __repr__(self):
        return "<LinkCrawlerJobsQuery>"

    def default():
        return LinkCrawlerJobsQuery(collectorInfo=True, jobIds=None)


class LinkQuery(BaseModel):

    addedDate: Optional[bool]
    bytesLoaded: Optional[bool]
    bytesTotal: Optional[bool]
    comment: Optional[bool]
    enabled: Optional[bool]
    eta: Optional[bool]
    extractionStatus: Optional[bool]
    finished: Optional[bool]
    finishedDate: Optional[bool]
    host: Optional[bool]
    jobUUIDs: Optional[List[int]]
    maxResults: Optional[int]
    packageUUIDs: Optional[List[int]]
    password: Optional[bool]
    priority: Optional[bool]
    running: Optional[bool]
    skipped: Optional[bool]
    speed: Optional[bool]
    startAt: Optional[int]
    status: Optional[bool]
    url: Optional[bool]

    def __repr__(self):
        return "<LinkQuery>"

    def default():
        return LinkQuery(
            addedDate=True,
            bytesLoaded=True,
            bytesTotal=True,
            comment=True,
            enabled=True,
            eta=True,
            extractionStatus=True,
            finished=True,
            finishedDate=True,
            host=True,
            jobUUIDs=None,
            maxResults=-1,
            packageUUIDs=None,
            password=True,
            priority=True,
            running=True,
            skipped=True,
            speed=True,
            startAt=0,
            status=True,
            url=True,
        )


class LogFolder(BaseModel):

    created: Optional[int]
    current: Optional[bool]
    lastModified: Optional[int]

    def __repr__(self):
        return "<LogFolder>"


MenuStructure = ForwardRef("MenuStructure")


class MenuStructure(BaseModel):

    children: Optional[List[MenuStructure]]
    icon: Optional[str]
    id: Optional[str]
    name: Optional[str]
    type: Optional[MenuType]

    def __repr__(self):
        return f"<MenuStructure ({self.menuId})>"


class PackageQuery(BaseModel):

    bytesLoaded: Optional[bool]
    bytesTotal: Optional[bool]
    childCount: Optional[bool]
    comment: Optional[bool]
    enabled: Optional[bool]
    eta: Optional[bool]
    finished: Optional[bool]
    hosts: Optional[bool]
    maxResults: Optional[int]
    packageUUIDs: Optional[List[int]]
    priority: Optional[bool]
    running: Optional[bool]
    saveTo: Optional[bool]
    speed: Optional[bool]
    startAt: Optional[int]
    status: Optional[bool]

    def __repr__(self):
        return "<PackageQuery>"

    def default():
        return PackageQuery(
            bytesLoaded=True,
            bytesTotal=True,
            childCount=True,
            comment=True,
            enabled=True,
            eta=True,
            finished=True,
            hosts=True,
            maxResults=-1,
            packageUUIDs=None,
            priority=True,
            running=True,
            saveTo=True,
            speed=True,
            startAt=0,
            status=True,
        )


class Plugin(BaseModel):

    abstractType: Optional[AbstractType]
    className: Optional[str]
    defaultValue: Optional[Any]
    displayName: Optional[str]
    docs: Optional[str]
    enumLabel: Optional[str]
    enumOptions: Optional[Any]
    interfaceName: Optional[str]
    key: Optional[str]
    pattern: Optional[str]
    storage: Optional[str]
    type: Optional[str]
    value: Optional[Any]
    version: Optional[str]

    def __repr__(self):
        return f"<Plugin ({self.className})>"


class PluginsQuery(BaseModel):

    pattern: Optional[str]
    version: Optional[str]

    def __repr__(self):
        return f"<PluginsQuery ({self.pattern})>"

    def default():
        return PluginsQuery(pattern="", version=None)


class PublisherResponse(BaseModel):

    eventids: Optional[List[str]]
    publisher: Optional[str]

    def __repr__(self):
        return f"<PublisherResponse ({self.publisher})>"


class SubscriptionResponse(BaseModel):

    exclusions: Optional[List[str]]
    maxKeepalive: Optional[int]
    maxPolltimeout: Optional[int]
    subscribed: Optional[bool]
    subscriptionid: Optional[int]
    subscriptions: Optional[List[str]]

    def __repr__(self):
        return f"<SubscriptionResponse ({self.subscription_id})>"


class IPandPort(BaseModel):

    port: int
    ip: str


class DirectConnectionInfos(BaseModel):

    infos: Optional[List[IPandPort]]
    rebindProtectionDetected: Optional[bool]
    mode: Optional[str]



from .jd_types import Extension, ExtensionQuery
from typing import Any, List


class Extensions:
    def __init__(self, device):
        self.device = device
        self.endpoint = "extensions"

    def action(self, route: str, params: Any = None) -> Any:
        route = f"/{self.endpoint}{route}"
        return self.device.connection_helper.action(route, params)

    def install(self, extension_id: str) -> bool:
        """Install the extension with extension_id.

        :param extension_id: The ID of the extension
        :type extension_id: str
        :return: Success
        :rtype: boolean
        """

        params = [extension_id]
        resp = self.action("/install", params)
        return resp

    def is_enabled(self, extension_id: str) -> bool:
        """Check if the extension of extension_id is enabled.

        :param extension_id: ID of the extension
        :type extension_id: str
        :return: Is enabled
        :rtype: boolean
        """

        params = [extension_id]
        resp = self.action("/isEnabled", params)
        return resp

    def is_installed(self, extension_id: str) -> bool:
        """Check if the extension of extension_id is installed.

        :param extension_id: ID of the extension
        :type extension_id: str
        :return: Is installed
        :rtype: boolean
        """

        params = [extension_id]
        resp = self.action("/isInstalled", params)
        return resp

    def list_extensions(
        self, query: ExtensionQuery = ExtensionQuery.default()
    ) -> List[Extension]:
        """List all extensions.

        :param query: A query to filter by (default: all)
        :type query: jd_types.ExtensionQuery
        :result: A list of extensions
        :rtype: List[jd_types.Extension]
        """

        params = [query.dict()]
        resp = list(self.action("/list", params))

        extensions = []
        for ext in resp:
            extension = Extension(**ext)
            extensions.append(extension)

        return extensions

    def set_enabled(self, extension_id: str, enabled: bool) -> bool:
        """Enable/Disable an extensions.

        :param extension_id: ID of the extension
        :type extension_id: str
        :param enabled: Enable or disable
        :type enabled: boolean
        :return: Success
        :rtype: boolean
        """

        params = [extension_id, enabled]
        resp = self.action("/setEnabled", params)
        return resp



from .jd_types import AdvancedConfigQuery, AdvancedConfigAPIEntry, Plugin, PluginsQuery
from typing import Any


class Plugins:
    def __init__(self, device):
        self.device = device
        self.endpoint = "plugins"

    def action(self, route: str, params: Any = None) -> Any:
        route = f"/{self.endpoint}{route}"
        return self.device.connection_helper.action(route, params)

    def get(self, interface_name, display_name, key):
        """Get a plugin."""

        params = [interface_name, display_name, key]
        resp = self.action("/get", params)

        return resp

    def get_all_plugin_regex(self):
        """Get all plugin regular expressions."""

        resp = self.action("/getAllPluginRegex")

        return resp

    def get_plugin_regex(self, url):
        """Get plugin regular expressions for a url."""

        params = [url]
        resp = self.action("/getPluginRegex", params)

        return resp

    def list(self, plugins_query=PluginsQuery.default()):
        """List plugins with query."""

        params = [plugins_query.dict()]
        resp = self.action("/list", params)

        plugins = []
        for p in resp:
            plugin = Plugin(**p)
            plugins.append(plugin)

        return plugins

    def query(self, config_query=AdvancedConfigQuery.default()):
        """Query plugin configurations."""

        params = [config_query.dict()]
        resp = self.action("/query", params)

        entries = []
        for c in resp:
            entry = AdvancedConfigAPIEntry(**c)
            entries.append(entry)

        return entries

    def reset(self, interface_name, display_name, key):
        """Reset plugin config."""

        params = [interface_name, display_name, key]
        resp = self.action("/reset", params)

        return resp

    def set(self, interface_name, display_name, key, new_value):
        """Set a plugin config value."""

        params = [interface_name, display_name, key, new_value]
        resp = self.action("/set", params)

        return resp



"""
pyjd
====

*pyjd* is a Python wrapper for the JDownloader API.

:copyright: (c) 2022 Philipp Glaum
:license: GPLv2, see LICENSE for more details
"""

__title__ = "pyjd"
__version__ = "2.0-alpha"
__author__ = "Philipp Glaum"
__copyright__ = "Copyright (c) 2022 Philipp Glaum"
__license__ = "GPLv2"



from .jd_types import (
    AddLinksQuery,
    CrawledLink,
    CrawledLinkQuery,
    CrawledPackage,
    CrawledPackageQuery,
    DeleteAction,
    JobLinkCrawler,
    LinkCollectingJob,
    LinkCrawlerJobsQuery,
    LinkVariant,
)
from typing import Any


class LinkGrabber:
    def __init__(self, device):
        self.device = device
        self.endpoint = "linkgrabberv2"

    def action(self, route: str, params: Any = None) -> Any:
        route = f"/{self.endpoint}{route}"
        return self.device.connection_helper.action(route, params)

    def abort(self, job_id: int = None) -> bool:
        """Abort one or all jobs.

        :param job_id: If this is given, only the job `job_id` will be aborted.
        :type job_id: int
        :return: Success
        :rtype: bool
        """

        params = None
        if job_id:
            params = [job_id]

        resp = self.action("/abort", params)
        return resp

    def add_container(self, container_type: str, content: str) -> LinkCollectingJob:
        """Add a container of type and content.

        :param container_type: The type of the container
        :type container_type: str
        :param content: The content of the container
        :type content: str
        :return: A link collecting job
        :rtype: LinkCollectingJob
        """

        params = [container_type, content]
        resp = self.action("/addContainer", params)
        job = LinkCollectingJob(**resp)
        return job

    def add_links(self, add_links_query: AddLinksQuery) -> LinkCollectingJob:
        """
        Add links to the linkcollector

        :param add_links_query: An AddLinksQuery object
        :type params: AddLinksQuery
        :return: A link collecting job
        :rtype: LinkCollectingJob
        """

        params = [add_links_query.dict()]
        resp = self.action("/addLinks", params)
        job = LinkCollectingJob(**resp)
        return job

    def add_variant_copy(
        self,
        link_id: int,
        destination_after_link_id: int,
        destination_package_id: int,
        variant_id: int,
    ) -> Any:
        """Unkown."""

        params = [
            link_id,
            destination_after_link_id,
            destination_package_id,
            variant_id,
        ]
        resp = self.action("/addVariantCopy", params)
        return resp

    def cleanup(
        self,
        delete_action: DeleteAction,
        mode,
        selection_type,
        link_ids=[],
        package_ids=[],
    ):
        """Clean packages and/or links of the linkgrabber list.

        Requires at least a ``package_ids`` or ``link_ids`` list, or both.

        :param package_ids: Package UUIDs.
        :type package_ids: List[int]
        :param link_ids: Link UUIDs.
        :type link_ids: List[int]
        :param action: DeleteAction to be done on cleanup.
        :type action: jd_types.DeleteAction
        :param mode: Mode to use.
        :type mode: jd_types.Mode
        :param selection_type: Type of selection to use.
        :type selection_type: jd_types.SelectionType
        """

        params = [link_ids, package_ids]
        params += [delete_action.value, mode.value, selection_type.value]
        resp = self.action("/cleanup", params)
        return resp

    def clear_list(self):
        """Clears Linkgrabbers list."""

        resp = self.action("/clearList")
        return resp

    def get_children_changed(self, structure_watermark):
        """Unkown."""

        params = [structure_watermark]
        resp = self.action("/getChildrenChanged", params)
        return resp

    def get_download_folder_history_selection_base(self):
        """Returns the download folder selection.

        :return: List of strings with paths to available download folders.
        :rtype: List[str]
        """

        resp = self.action("/getDownloadFolderHistorySelectionBase")
        return resp

    def get_download_urls(self, link_ids, package_ids, url_display_types=["ORIGIN"]):
        """Gets download urls from Linkgrabber.

        :param package_ids: Package UUIDs.
        :type package_ids: List[int]
        :param link_ids: link UUIDs.
        :type link_ids: List[int]
        :param url_display_types: No clue. Not documented
        :type url_display_types: List
        """

        params = [link_ids, package_ids, url_display_types]
        resp = self.action("/getDownloadUrls", params)
        return resp

    def get_package_count(self):
        """Get package count in linkgrabber"""

        resp = self.action("/getPackageCount")
        return resp

    def get_variants(self, link_id):
        """Gets the variants of a url/download (not package)

        For example a youtube link gives you a package with three downloads,
        the audio, the video and a picture, and each of those downloads have
        different variants (audio quality, video quality, and picture quality).

        :param params: The UUID of the download you want the variants.
        :type params: List[int]
        :rtype: Variants in a list with dictionaries like this one:

            .. code-block :: json

                [
                    {
                        "id": "M4A_256",
                        "name": "256kbit/s M4A-Audio"
                    },
                    {
                        "id": "AAC_256",
                        "name": "256kbit/s AAC-Audio"
                    }
                ]

        """

        params = [link_id]
        resp = self.action("/getVariants", params)

        link_variants = []
        for variant in resp:
            link_variant = LinkVariant(variant)
            link_variants.append(link_variant)

        return link_variants

    def is_collecting(self):
        """Bool status query about the collecting process"""

        resp = self.action("/isCollecting")
        return resp

    def move_links(self, link_ids, after_link_id, dest_package_id):
        """Unkown."""

        params = [link_ids, after_link_id, dest_package_id]
        resp = self.action("/moveLinks", params)
        return resp

    def move_packages(self, package_ids, after_dest_package_id):
        """Unkown."""

        params = [package_ids, after_dest_package_id]
        resp = self.action("/movePackages", params)
        return resp

    def move_to_downloadlist(self, link_ids, package_ids):
        """Moves packages and/or links to download list.

        :param link_ids: Link UUIDs.
        :type link_ids: List[int]
        :param package_ids: Package UUIDs.
        :type package_ids: List[int]
        """

        params = [link_ids, package_ids]
        resp = self.action("/moveToDownloadlist", params)
        return resp

    def move_to_new_package(self, link_ids, package_ids, new_pkg_name, download_path):
        """Moves packages and/or links to a new package

        :param link_ids: Link UUIDs.
        :type link_ids: List[int]
        :param package_ids: Package UUIDs.
        :type package_ids: List[int]
        :param new_pkg_name: The name of the new package
        :type new_pkg_name: str
        :param download_path: Download path for the new package
        :type download_path: str
        """

        params = [link_ids, package_ids, new_pkg_name, download_path]
        resp = self.action("/movetoNewPackage", params)
        return resp

    def query_link_crawler_jobs(
        self, link_crawler_jobs_query=LinkCrawlerJobsQuery.default()
    ):
        """Query link crawler jobs.

        :param crawled_link_query: Query to filter by
        :type crawled_link_query: jd_types.LinkCrawlerJobsQuery
        :return: List of crawled packages
        :rtype: List[jd_types.JobLinkCrawler]
        """

        params = [link_crawler_jobs_query.dict()]
        resp = self.action("/queryLinkCrawlerJobs", params)

        crawler_jobs = []
        for job in resp:
            crawler_job = JobLinkCrawler(**job)
            crawler_jobs.append(crawler_job)

        return crawler_jobs

    def query_links(self, crawled_link_query=CrawledLinkQuery.default()):
        """Get the links in the linkcollector/linkgrabber

        :param params: A CrawledLinkQuery object with options.
        :type params: CrawledLinkQuery
        :return: List of CrawledLink objects
        :rtype: List[CrawledLink]

        """

        params = [crawled_link_query.dict()]
        resp = self.action("/queryLinks", params)

        crawled_links = []
        for link in resp:
            crawled_link = CrawledLink(**link)
            crawled_links.append(crawled_link)

        return crawled_links

    def query_packages(self, crawled_package_query=CrawledPackageQuery.default()):
        """Get the crawled packages in the linkgrabber

        :param params: A dictionary of parameters to pass.
        :type params: jd_types.CrawledPackageQuery
        :return: A list of crawled packages:
        """

        params = [crawled_package_query.dict()]
        resp = self.action("/queryPackages", params)

        crawled_packages = []
        for package in resp:
            crawled_package = CrawledPackage(**package)
            crawled_packages.append(crawled_package)

        return crawled_packages

    def remove_links(self, link_ids, package_ids):
        """Unknown."""

        params = [link_ids, package_ids]
        resp = self.action("/removeLinks", params)
        return resp

    def rename_link(self, link_id, new_name):
        """Unknown."""

        params = [link_id, new_name]
        resp = self.action("/renameLink", params)
        return resp

    def rename_package(self, package_id, new_name):
        """Unknown."""

        params = [package_id, new_name]
        resp = self.action("/renamePackage", params)
        return resp

    def set_download_directory(self, directory, package_ids):
        """Unknown."""

        params = [directory, package_ids]
        resp = self.action("/setDownloadDirectory", params)
        return resp

    def set_download_password(self, link_ids, package_ids, password):
        """Unknown."""

        params = [link_ids, package_ids, password]
        resp = self.action("/setDownloadPassword", params)
        return resp

    def set_enabled(self, enabled, link_ids, package_ids):
        """Sets the UUIDs as enabled

        :param enabled: Enable or disable the IDs
        :type enabled: bool
        :param link_ids: List of link UUIDs
        :type link_ids: List[int]
        :param package_ids: List of package UUIDs
        :type package_ids: List[int]
        """

        params = [enabled, link_ids, package_ids]
        resp = self.action("/setEnabled", params=params)
        return resp

    def set_priority(self, priority, link_ids, package_ids):
        """Sets the priority of links or packages.

        :param package_ids: Package UUIDs.
        :type package_ids: List[int]
        :param link_ids: link UUIDs.
        :type link_ids: List[int]
        :param priority: Priority to set.
            Priorities: HIGHEST, HIGHER, HIGH, DEFAULT, LOWER
        :type priority: str
        """

        params = [priority, link_ids, package_ids]
        resp = self.action("/setPriority", params)
        return resp

    def set_variant(self, link_id, variant_id):
        """Unknown."""

        params = [link_id, variant_id]
        resp = self.action("/setVariant", params)
        return resp

    def split_package_by_hoster(self, link_ids, pkg_ids):
        """Unknown."""

        params = [link_ids, pkg_ids]
        resp = self.action("/splitPackageByHoster", params)
        return resp

    def start_online_check(self, link_ids, package_ids):
        """Unknown."""

        params = [link_ids, package_ids]
        resp = self.action("/startOnlineStatusCheck", params)
        return resp

    def help(self):
        """Returns the API help."""

        resp = self.action("/linkgrabberv2/help", http_action="GET")
        return resp



from .jd_types import IconDescriptor
from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    from .jd_device import JDDevice


class Content:
    def __init__(self, device: "JDDevice") -> None:
        self.device = device
        self.endpoint = "contentV2"

    def action(self, route: str, params: Any = None, binary: bool = False) -> Any:
        route = f"/{self.endpoint}{route}"
        return self.device.connection_helper.action(route, params, binary=binary)

    def get_fav_icon(self, hostername: str) -> bytes:
        """Get the fav icon for a hoster.

        :param hostername: Name of the hoster for which the favicon will be
            returned
        :type hostername: str
        :returns: The favicon as png
        :rtype: bytes
        """

        params = [hostername]
        resp = self.action("/getFavIcon", params, True)
        return resp

    def get_file_icon(self, filename: str) -> bytes:
        """Get the file icon.

        :param filename: The name of the icon
        :type filename: str
        :returns: The icon as png
        :rtype: bytes
        """

        params = [filename]
        resp = self.action("/getFileIcon", params, True)
        return resp

    def get_icon(self, key: str, size: int) -> bytes:
        """Get an icon, scaled for size.

        :param filename: The name of the icon
        :type filename: str
        :param size: The size of the icon in px (it's a square)
        :type size: int
        :returns: The icon as png
        :rtype: bytes
        """

        params = [key, size]
        resp = self.action("/getIcon", params, True)
        return resp

    def get_icon_description(self, key: str) -> IconDescriptor:
        """Get an icon description.

        :param key: The icon key
        :type key: str
        :returns: Description for the key
        :rtype: str
        """

        params = [key]
        resp = self.action("/getIconDescription", params)
        descriptor = IconDescriptor(**resp)
        return descriptor



from .jd_types import Context, MenuStructure
from typing import Any, List, TYPE_CHECKING

if TYPE_CHECKING:
    from .jd_device import JDDevice


class UI:
    def __init__(self, device: "JDDevice") -> None:
        self.device = device
        self.endpoint = "ui"

    def action(self, route: str, params: Any = None) -> Any:
        route = f"/{self.endpoint}{route}"
        return self.device.connection_helper.action(route, params)

    def get_menu(self, context: Context) -> MenuStructure:
        """Get the custom menu structure for the desired context."""

        params = [context.value]
        resp = self.action("/getMenu", params)
        menu_item = MenuStructure(**resp)
        return menu_item

    def invoke_action(
        self,
        context: Context,
        action_id: int,
        link_ids: List[int],
        package_ids: List[int],
    ) -> str:
        """Invoke a menu action on our selection and get the results."""

        params = [context, action_id, link_ids, package_ids]
        resp = self.action("/invokeAction", params)
        return resp
